class AppFileLib:
    # Format to: app_name
    default_config_toml = """\
# Flask-BigApp Config File
# ------------------------
# Updates the Flask app config with the variables below.
# If any variable below does not exist in the standard Flask env
# vars it is created and will be accessible using
# app.config. All key names defined below will be
# capitalised when imported.
[FLASK]
DEBUG = false
#PROPAGATE_EXCEPTIONS = true
TRAP_HTTP_EXCEPTIONS = false
#TRAP_BAD_REQUEST_ERRORS = true
SECRET_KEY = "super_secret_key"
SESSION_COOKIE_NAME = "session"
#SESSION_COOKIE_DOMAIN = "domain-here.com"
#SESSION_COOKIE_PATH = "/"
SESSION_COOKIE_HTTPONLY = true
SESSION_COOKIE_SECURE = false
SESSION_COOKIE_SAMESITE = "Lax"
PERMANENT_SESSION_LIFETIME = 3600  # 1 hour
SESSION_REFRESH_EACH_REQUEST = true
USE_X_SENDFILE = false
#SEND_FILE_MAX_AGE_DEFAULT = 43200
ERROR_404_HELP = true
#SERVER_NAME = "localhost:5000"
APPLICATION_ROOT = "/"
PREFERRED_URL_SCHEME = "http"
#MAX_CONTENT_LENGTH = 0
#TEMPLATES_AUTO_RELOAD = true
EXPLAIN_TEMPLATE_LOADING = false
MAX_COOKIE_SIZE = 4093


# This will set the default session variables for the app.
# Anything here will be accessible using session["your_var_name"]
# or session.get("your_var_name")
[SESSION]
logged_in = false

# These settings are spcific to the Flask-SQLAlchemy extension.
# Anything here will be accessible using app.config
[SQLALCHEMY]
SQLALCHEMY_ECHO = false
SQLALCHEMY_TRACK_MODIFICATIONS = false
SQLALCHEMY_RECORD_QUERIES = false

# [DATABASE.MAIN] is loaded as SQLALCHEMY_DATABASE_URI
# Dialets = mysql / postgresql / sqlite / oracle / mssql

# Uncomment below to generate the SQLALCHEMY_DATABASE_URI.
#[DATABASE.MAIN]
#ENABLED = true
#DIALECT = "sqlite"
#DATABASE_NAME = "database"
#LOCATION = "db"
#PORT = ""
#USERNAME = "database"
#PASSWORD = "password"

# Adding another database is as simple as adding a new section.
# [DATABASE.ANOTHER] will then be accessible using SQLALCHEMY_BINDS
# The bind key will be stored as a lowercase value, so "ANOTHER" will
# be accessible as "another"
# You can then use the bind key in the model as follows:
# class MyModel(db.Model):
#     __bind_key__ = "another"
#     ...

# Uncomment below to generate and add to SQLALCHEMY_BINDS.
#[DATABASE.ANOTHER]
#ENABLED = true
#DIALECT = "sqlite"
#DATABASE_NAME = "database_another"
#LOCATION = "db"
#PORT = ""
#USERNAME = "user"
#PASSWORD = "password"
"""

    init_py = """\
from flask import Flask
from {app_name}.extensions import bigapp, db


def create_app():
    app = Flask(__name__)
    bigapp.init_app(app)
    #db.init_app(app)

    bigapp.import_app_resources(
        app_factories=["collection"]
    )

    bigapp.import_blueprints("blueprints")

    #bigapp.import_models("models")

    return app
"""

    extensions_init_py = """\
from flask_bigapp import BigApp
from flask_sqlalchemy import SQLAlchemy

bigapp = BigApp()
db = SQLAlchemy()
"""

    # Format to: app_name
    models_init_py = """\
from sqlalchemy import select, update, delete, insert
from sqlalchemy.orm import relationship

from {app_name}.extensions import db

__all__ = [
    "db",
    "select",
    "update",
    "delete",
    "insert",
    "relationship",
]
"""

    # Format to: None
    models_example_table_py = """\
from . import *


class ExampleUserTable(db.Model):
    user_id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(256), nullable=False)
    password = db.Column(db.String(512), nullable=False)
    salt = db.Column(db.String(4), nullable=False)
    private_key = db.Column(db.String(256), nullable=False)
    disabled = db.Column(db.Boolean)

    @classmethod
    def get_by_id(cls, user_id):
        return db.session.execute(
            select(cls).filter_by(user_id=user_id).limit(1)
        ).scalar_one_or_none()

    @classmethod
    def create(cls, username, password, disabled):
        from flask_bigapp.auth import Auth

        salt = Auth.generate_salt()
        salt_pepper_password = Auth.hash_password(password, salt)
        private_key = Auth.generate_private_key(username)

        db.session.execute(
            insert(cls).values(
                username=username,
                password=salt_pepper_password,
                salt=salt,
                private_key=private_key,
                disabled=disabled,
            )
        )
        db.session.commit()

    @classmethod
    def update(cls, user_id, username, private_key, disabled):
        db.session.execute(
            update(cls).where(
                cls.user_id == user_id  # noqa
            ).values(
                username=username,
                private_key=private_key,
                disabled=disabled,
            )
        )
        db.session.commit()

    @classmethod
    def delete(cls, user_id):
        db.session.execute(
            delete(cls).where(
                cls.user_id == user_id  # noqa
            )
        )
        db.session.commit()
"""



from pathlib import Path
import click

from .helpers import to_snake_case
from .helpers import Sprinkles as Sp
from .filelib import BlueprintFileLib
from .filelib import InitAppBlueprintFileLib


def init_new_app_blueprint(folder, name):
    cwd = Path.cwd()
    if folder != "Current Working Directory":
        cwd = Path(cwd / folder)
    if not cwd.exists():
        click.echo(
            f"{Sp.FAIL}{folder} does not exist.{Sp.END}")
        return

    name = to_snake_case(name)

    # Prepare blueprint folder structure
    bp_folder = cwd / name
    bp_routes_folder = bp_folder / "routes"
    bp_templates_folder = bp_folder / "templates" / name

    # Prepare blueprint files
    bp_init_py = bp_folder / "__init__.py"
    bp_config_toml = bp_folder / "config.toml"
    bp_routes_index_py = bp_routes_folder / "index.py"
    bp_templates_index_html = bp_templates_folder / "index.html"

    # Prepare blueprint folders for loop creation
    folders = (
        bp_folder,
        bp_routes_folder,
        bp_templates_folder,
    )

    # Loop create folders
    for folder in folders:
        if not folder.exists():
            folder.mkdir(parents=True)
            click.echo(f"{Sp.OKGREEN}Blueprint folder: {folder.name}, created{Sp.END}")
        else:
            click.echo(f"{Sp.WARNING}Blueprint folder already exists: {folder.name}, skipping{Sp.END}")

    # Create __init__.py
    if not bp_init_py.exists():
        bp_init_py.write_text(BlueprintFileLib.init_py, encoding="utf-8")
        click.echo(f"{Sp.OKGREEN}Blueprint __init__ created{Sp.END}")
    else:
        click.echo(f"{Sp.WARNING}Blueprint __init__ already exists, skipping{Sp.END}")

    # Create config.toml
    if not bp_config_toml.exists():
        bp_config_toml.write_text(
            InitAppBlueprintFileLib.config_toml.format(
                name=name,
                url_prefix="",
            ), encoding="utf-8"
        )
        click.echo(f"{Sp.OKGREEN}Blueprint config, created{Sp.END}")
    else:
        click.echo(f"{Sp.WARNING}Blueprint config already exists, skipping{Sp.END}")

    # Create blueprint index.py route
    if not bp_routes_index_py.exists():
        bp_routes_index_py.write_text(
            BlueprintFileLib.routes_index_py, encoding="utf-8")
        click.echo(f"{Sp.OKGREEN}Blueprint route: {bp_routes_index_py.name}, created{Sp.END}")
    else:
        click.echo(f"{Sp.WARNING}Blueprint route already exists: {bp_routes_index_py.name}, skipping{Sp.END}")

    # Create blueprint index.html template
    if not bp_templates_index_html.exists():
        bp_templates_index_html.write_text(
            InitAppBlueprintFileLib.templates_index_html.format(name=name), encoding="utf-8")
        click.echo(f"{Sp.OKGREEN}Blueprint template file: {bp_templates_index_html.name}, created{Sp.END}")
    else:
        click.echo(
            f"{Sp.WARNING}Blueprint template file already exists: {bp_templates_index_html.name}, skipping{Sp.END}")

    click.echo(f"{Sp.OKGREEN}Blueprint created: {bp_folder}{Sp.END}")



from dataclasses import dataclass


@dataclass
class Resources:
    # Format to: secret_key
    default_config = """\
# Flask-BigApp Config File
# ------------------------
# Updates the Flask app config with the variables below.
# If any variable below does not exist in the standard Flask env
# vars it is created and will be accessible using
# app.config. All key names defined below will be
# capitalised when imported.
[FLASK]
DEBUG = false
#PROPAGATE_EXCEPTIONS = true
TRAP_HTTP_EXCEPTIONS = false
#TRAP_BAD_REQUEST_ERRORS = true
SECRET_KEY = "super_secret_key"
SESSION_COOKIE_NAME = "session"
#SESSION_COOKIE_DOMAIN = "domain-here.com"
#SESSION_COOKIE_PATH = "/"
SESSION_COOKIE_HTTPONLY = true
SESSION_COOKIE_SECURE = false
SESSION_COOKIE_SAMESITE = "Lax"
PERMANENT_SESSION_LIFETIME = 3600  # 1 hour
SESSION_REFRESH_EACH_REQUEST = true
USE_X_SENDFILE = false
#SEND_FILE_MAX_AGE_DEFAULT = 43200
ERROR_404_HELP = true
#SERVER_NAME = "localhost:5000"
APPLICATION_ROOT = "/"
PREFERRED_URL_SCHEME = "http"
#MAX_CONTENT_LENGTH = 0
#TEMPLATES_AUTO_RELOAD = true
EXPLAIN_TEMPLATE_LOADING = false
MAX_COOKIE_SIZE = 4093


# This will set the default session variables for the app.
# Anything here will be accessible using session["your_var_name"]
# or session.get("your_var_name")
[SESSION]
logged_in = false

# These settings are spcific to the Flask-SQLAlchemy extension.
# Anything here will be accessible using app.config
[SQLALCHEMY]
SQLALCHEMY_ECHO = false
SQLALCHEMY_TRACK_MODIFICATIONS = false
SQLALCHEMY_RECORD_QUERIES = false

# [DATABASE.MAIN] is loaded as SQLALCHEMY_DATABASE_URI
# Dialets = mysql / postgresql / sqlite / oracle / mssql

# Uncomment below to generate the SQLALCHEMY_DATABASE_URI.
#[DATABASE.MAIN]
#ENABLED = true
#DIALECT = "sqlite"
#DATABASE_NAME = "database"
#LOCATION = "db"
#PORT = ""
#USERNAME = "database"
#PASSWORD = "password"

# Adding another database is as simple as adding a new section.
# [DATABASE.ANOTHER] will then be accessible using SQLALCHEMY_BINDS
# The bind key will be stored as a lowercase value, so "ANOTHER" will
# be accessible as "another"
# You can then use the bind key in the model as follows:
# class MyModel(db.Model):
#     __bind_key__ = "another"
#     ...

# Uncomment below to generate and add to SQLALCHEMY_BINDS.
#[DATABASE.ANOTHER]
#ENABLED = true
#DIALECT = "sqlite"
#DATABASE_NAME = "database_another"
#LOCATION = "db"
#PORT = ""
#USERNAME = "user"
#PASSWORD = "password"
"""



import re


def to_snake_case(string):
    """
    Thank you openai
    """
    # Replace any non-alphanumeric characters with underscores
    string = re.sub(r'[^a-zA-Z0-9]', '_', string)
    # Remove any consecutive underscores
    string = re.sub(r'_{2,}', '_', string)
    # Convert the string to lowercase
    string = string.lower()
    return string


class Sprinkles:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'



import typing as t
from functools import wraps

from flask import flash, abort
from flask import redirect
from flask import session
from flask import url_for


def _check_against_values_allowed(
        session_value: t.Union[list, str, int, bool],
        values_allowed: t.Union[t.List[t.Union[str, int, bool]], str, int, bool],
) -> bool:
    if isinstance(values_allowed, list):
        if isinstance(session_value, list):
            for value in session_value:
                if value in values_allowed:
                    return True
            return False

        if session_value in values_allowed:
            return True
        return False

    if session_value == values_allowed:
        return True

    return False


def login_check(
        session_key: str,
        values_allowed: t.Union[t.List[t.Union[str, int, bool]], str, int, bool],
        fail_endpoint: t.Optional[str] = None,
        pass_endpoint: t.Optional[str] = None,
        endpoint_kwargs: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
        message: t.Optional[str] = None,
        message_category: str = "message"
):
    """
    A decorator that checks if the specified session key exists and contains the specified value.

    Example of a route that requires a user to be logged in:

    @bp.route("/admin", methods=["GET"])
    @login_check('logged_in', True, 'blueprint.login_page', message="Login needed")
    def admin_page():
        ...

    Example of a route that if the user is already logged in, redirects to the specified endpoint:

    @bp.route("/login-page", methods=["GET"])
    @login_check('logged_in', True, 'blueprint.admin_page', message="Already logged in")
    def login_page():
        ...

    :param session_key: The session key to check for.
    :param values_allowed: A list of or singular value(s) that the session key must contain.
    :param fail_endpoint: The endpoint to redirect to if the session key does not exist or
                          match the pass_value.
    :param pass_endpoint: The endpoint to redirect to if the session key passes.
                          Used to redirect away from login pages, if already logged in.
    :param endpoint_kwargs: A dictionary of keyword arguments to pass to the redirect endpoint.
    :param message: If a message is specified, a flash message is shown.
    :param message_category: The category of the flash message.
    """

    def login_check_wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            skey = session.get(session_key)

            if skey is None:
                if fail_endpoint:
                    if endpoint_kwargs:
                        return redirect(url_for(fail_endpoint, **endpoint_kwargs))

                    return redirect(url_for(fail_endpoint))

                return func(*args, **kwargs)

            if skey is not None:
                if _check_against_values_allowed(skey, values_allowed):
                    if pass_endpoint:
                        if message:
                            flash(message, message_category)

                        if endpoint_kwargs:
                            return redirect(url_for(pass_endpoint, **endpoint_kwargs))

                        return redirect(url_for(pass_endpoint))

                    return func(*args, **kwargs)

                if fail_endpoint:
                    if endpoint_kwargs:
                        return redirect(url_for(fail_endpoint, **endpoint_kwargs))

                    return redirect(url_for(fail_endpoint))

                return func(*args, **kwargs)

            return abort(403)

        return inner

    return login_check_wrapper


def permission_check(
        session_key: str,
        values_allowed: t.Union[t.List[t.Union[str, int, bool]], str, int, bool],
        fail_endpoint: t.Optional[str] = None,
        endpoint_kwargs: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
        message: t.Optional[str] = None,
        message_category: str = "message"
):
    """
    A decorator that checks if the specified session key exists and its value(s) match the specified value(s).

    Example:

    @bp.route("/admin-page", methods=["GET"])
    @login_check('logged_in', 'blueprint.login_page') <- can be mixed with login_check
    @permission_check('permissions', ['admin'], 'www.index', message="Failed message", message_category="error")
    def admin_page():
        ...

    :param session_key: The session key to check for.
    :param values_allowed: A list of or singular value(s) that the session key must contain.
    :param fail_endpoint: The endpoint to redirect to if the
                          session key does not exist or does not contain the
                          specified values.
    :param endpoint_kwargs: A dictionary of keyword arguments to pass to the redirect endpoint.
    :param message: If a message is specified, a flash message is shown.
    :param message_category: The category of the flash message.
    """

    def permission_check_wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            skey = session.get(session_key)

            if skey:
                if _check_against_values_allowed(skey, values_allowed):
                    return func(*args, **kwargs)

            if message:
                flash(message, message_category)

            if fail_endpoint:
                if endpoint_kwargs:
                    return redirect(url_for(fail_endpoint, **endpoint_kwargs))

                return redirect(url_for(fail_endpoint))

            return abort(403)

        return inner

    return permission_check_wrapper


def api_login_check(
        session_key: str,
        values_allowed: t.Union[t.List[t.Union[str, int, bool]], str, int, bool],
        fail_json: t.Optional[t.Dict[str, t.Any]] = None
):
    """
    A decorator that is used to secure API routes that return a JSON response.

    Example of a route that requires a user to be logged in:

    @bp.route("/api/resource", methods=["GET"])
    @api_login_check('logged_in', True)
    def api_page():
        ...


    You can also supply your own failed return JSON:

    @bp.route("/api/resource", methods=["GET"])
    @api_login_check('logged_in', True, fail_json={"error": "You are not logged in."})
    def api_page():
        ...


    Default json_to_return is {"error": "You are not logged in."}

    :param session_key: The session key to check for.
    :param values_allowed: A list of or singular value(s) that the session key must contain.
    :param fail_json: JSON that is returned on failure.
    """

    def api_login_check_wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            skey = session.get(session_key)
            if skey:
                if _check_against_values_allowed(skey, values_allowed):
                    return func(*args, **kwargs)
            else:
                if fail_json:
                    return fail_json or {"error": "You are not logged in."}

        return inner

    return api_login_check_wrapper



import click

from .helpers import Sprinkles as Sp
from .blueprint import add_blueprint as _add_blueprint
from .init_new_app import init_new_app as _init_new_app


@click.group()
def cli():
    pass  # Entry Point


@cli.command("blueprint", help="Create a flask-bigapp blueprint")
@click.option(
    '-f', '--folder',
    nargs=1,
    default="Current Working Directory",
    prompt=(f'\n{Sp.WARNING}(Creation is relative to the current working directory){Sp.END}\n'
            f'Folder to create blueprint in'),
    help='The from_folder to create the blueprint in, defaults to the current working directory'
)
@click.option(
    '-n', '--name',
    nargs=1,
    default="my_new_blueprint",
    prompt='Name of the blueprint to create',
    help='The name of the blueprint to create'
)
def add_blueprint(folder, name):
    _add_blueprint(folder, name)


@cli.command("init", help="Create a new flask-bigapp app")
@click.option(
    '-n', '--name',
    nargs=1,
    default="app",
    prompt='What would you like to call your app?',
    help='The name of the app folder that will be created'
)
def init_new_app(name):
    _init_new_app(name)



import logging
import os
import sys
from importlib import import_module
from inspect import getmembers
from inspect import isclass
from pathlib import Path
from typing import Dict, Union, Optional, List

from flask import Flask
from flask import Blueprint
from flask import session
from flask_sqlalchemy.model import DefaultMeta

from .helpers import init_app_config
from .registeries import ModelRegistry
from .utilities import cast_to_import_str


class BigApp:
    _app: Flask
    _app_name: str
    _app_path: Path
    _app_folder: Path
    _app_resources_imported: bool = False

    __model_registry__: ModelRegistry

    config_path: Path
    config: Dict

    def __init__(
            self,
            app: Optional[Flask] = None,
            app_config_file: Optional[str] = None,
            ignore_missing_env_variables: bool = False
    ) -> None:

        if app is not None:
            self.init_app(
                app,
                app_config_file,
                ignore_missing_env_variables
            )

    def init_app(
            self,
            app: Flask,
            app_config_file: Optional[str] = os.environ.get("BA_CONFIG"),
            ignore_missing_env_variables: bool = False
    ) -> None:

        """
        Initializes the flask app to work with flask-bigapp.

        :param app: The flask app to initialize.
        :param app_config_file: The config file to use.
        If not given will attempt to read from the environment.
        If no environment variable is found will use default.config.toml, which will be created if not found.
        :param ignore_missing_env_variables: If set to True will ignore missing environment variables.
        """

        if app is None:
            raise ImportError(
                "No app was passed in, do ba = BigApp(flaskapp) or app.init_app(flaskapp)")
        if not isinstance(app, Flask):
            raise TypeError(
                "The app that was passed in is not an instance of Flask")
        if app_config_file is None:
            app_config_file = "default.config.toml"

        self._app = app
        self._app_name = app.name
        self._app_path = Path(self._app.root_path)
        self._app_folder = self._app_path.parent
        self.config_path = self._app_path / app_config_file

        self.__model_registry__ = ModelRegistry()

        self.config = init_app_config(
            self.config_path,
            ignore_missing_env_variables,
            self._app
        )

    def import_app_resources(
            self,
            folder: str = "global",
            app_factories: Optional[List] = None,
            static_folder: str = "static",
            templates_folder: str = "templates",
            scope_root_folders_to: Optional[List] = None,
            scope_root_files_to: Optional[List] = None,
    ) -> None:
        """
        Import standard app resources from a single folder.

        :param folder: The folder to import from.
        Must be relative.
        :param app_factories: A list of function names to call with the app instance.
        ["collection"] => collection(app) will be called
        :param static_folder: The name of the static folder (if not found will be set to None)
        :param templates_folder: The name of the templates folder (if not found will be set to None)
        :param scope_root_folders_to: A list of folders to scope the import to
        ["cli", "routes"] => will only import from folder/cli/*.py and folder/routes/*.py
        :param scope_root_files_to: A list of files to scope the import to
        ["cli.py", "routes.py"] => will only import from folder/cli.py and folder/routes.py
        """

        if app_factories is None:
            app_factories = []

        if scope_root_folders_to is None:
            scope_root_folders_to = []

        if scope_root_files_to is None:
            scope_root_files_to = []

        if self._app_resources_imported:
            raise ImportError("The app resources can only be imported once.")

        self._app_resources_imported = True

        def process_module(import_location: str) -> tuple:
            module_file = import_module(import_location)
            flask_instance = True if [
                name for name, value in getmembers(module_file) if
                isinstance(value, Flask)
            ] else False

            return module_file, flask_instance

        skip_folders = ("static", "templates",)
        global_collection_folder = self._app_path / folder
        app_static_folder = global_collection_folder / static_folder
        app_templates_folder = global_collection_folder / templates_folder

        if not global_collection_folder.exists():
            raise ImportError(
                f"Cannot find global collection folder at {global_collection_folder}")

        if not global_collection_folder.is_dir():
            raise ImportError(
                f"Global collection must be a folder {global_collection_folder}")

        self._app.static_folder = app_static_folder.as_posix() if app_static_folder.exists() else None
        self._app.template_folder = app_templates_folder.as_posix() if app_templates_folder.exists() else None

        with self._app.app_context():
            for item in global_collection_folder.iterdir():

                if item.is_dir() and item.name not in skip_folders:
                    if scope_root_folders_to:
                        if item.name not in scope_root_folders_to:
                            continue

                    for py_file in item.glob("*.py"):
                        module, flask_instance_found = process_module(
                            f"{cast_to_import_str(self._app_name, item)}.{py_file.stem}")

                        for instance_factory in app_factories:
                            if hasattr(module, instance_factory):
                                getattr(module, instance_factory)(self._app)

                        if not flask_instance_found and not app_factories:
                            del sys.modules[module.__name__]

                if item.is_file() and item.suffix == ".py":
                    if scope_root_files_to:
                        if item.name not in scope_root_files_to:
                            continue

                    module, flask_instance_found = process_module(
                        cast_to_import_str(self._app_name, item))

                    for instance_factory in app_factories:
                        if hasattr(module, instance_factory):
                            getattr(module, instance_factory)(self._app)

                    if not flask_instance_found and not app_factories:
                        del sys.modules[module.__name__]

    def init_session(self) -> None:
        """
        Initialize the session variables found in the config.
        Use this method in the before_request route.
        """
        if self.config.get("SESSION"):
            for key, value in self.config.get("SESSION", {}).items():
                if key not in session:
                    session[key] = value

    def import_blueprint(self, blueprint: str) -> None:
        """
        Imports a single blueprint from the given path.

        :param blueprint: The blueprint (Python Package) to import.
        Must be relative.
        """
        if Path(blueprint).is_absolute():
            potential_bp = Path(blueprint)
        else:
            potential_bp = Path(self._app_path / blueprint)

        if potential_bp.exists() and potential_bp.is_dir():
            try:
                module = import_module(cast_to_import_str(self._app_name, potential_bp))
                for name, value in getmembers(module):
                    if isinstance(value, Blueprint):
                        if hasattr(value, "enabled"):
                            if value.enabled:
                                self._app.register_blueprint(value)
                            else:
                                logging.debug(f"Blueprint {name} is disabled")
                        else:
                            self._app.register_blueprint(value)
            except Exception as e:
                raise ImportError(f"Error when importing {potential_bp.name}: {e}")

    def import_blueprints(self, folder: str) -> None:
        """
        Imports all the blueprints in the given folder.

        :param folder: The folder to import from.
        Must be relative.
        """

        folder_path = Path(self._app_path / folder)

        for potential_bp in folder_path.iterdir():
            self.import_blueprint(potential_bp.as_posix())

    def import_models(
            self,
            file_or_folder: str
    ) -> None:
        """
        Imports all the models in the given file or folder.

        :param file_or_folder: The file or folder to import from.
        Must be relative.
        """

        def model_processor(path: Path):
            """
            Picks apart the model from_file and builds a registry of the models found.
            """
            import_string = cast_to_import_str(self._app_name, path)
            try:
                model_module = import_module(import_string)
                for name, value in getmembers(model_module, isclass):
                    if hasattr(value, "__tablename__"):
                        self.__model_registry__.add(name, value)

            except ImportError as e:
                raise ImportError(f"Error when importing {import_string}: {e}")

        if Path(file_or_folder).is_absolute():
            file_or_folder_path = Path(file_or_folder)
        else:
            file_or_folder_path = Path(self._app_path / file_or_folder)

        if file_or_folder_path.is_file() and file_or_folder_path.suffix == ".py":
            model_processor(file_or_folder_path)

        elif file_or_folder_path.is_dir():
            for model_file in [_ for _ in file_or_folder_path.iterdir() if "__" not in _.name]:
                model_processor(model_file)

    def model(self, class_: str) -> DefaultMeta:
        """
        Returns the model class for the given ORM class name

        bigapp.model("User") => <class 'app.models.User'>
        """
        return self.__model_registry__.class_(class_)

    def model_meta(self, class_: Union[str, DefaultMeta]) -> dict:
        """
        Returns meta information for the given ORM class name
        """

        def check_for_table_name(model_):
            if not hasattr(model_, "__tablename__"):
                raise AttributeError(f"{model_} is not a valid model")

        if isinstance(class_, str):
            model = self.__model_registry__.class_(class_)
            check_for_table_name(model)
            return {
                "location": model.__module__,
                "table_name": model.__tablename__,
            }

        return {
            "location": class_.__module__,
            "table_name": class_.__tablename__,
        }



class GlobalFileLib:
    # Format to: app_name
    collections_cli_py = """\
def collection(app: object):
    from {app_name}.extensions import db
    from {app_name}.models.example__table import ExampleUserTable

    @app.cli.command("create-tables")
    def create_tables():
        db.create_all()

    @app.cli.command("get-example-user")
    def get_example_user():
        result = ExampleUserTable.get_by_id(1)
        if not result:
            print("User not found.")
            return
        print(
            f\"\"\"
            user_id: {{result.user_id}}
            username: {{result.username}}
            salt: {{result.salt}}
            password: {{result.password}}
            private_key: {{result.private_key}}
            disabled: {{result.disabled}}
            \"\"\"
        )

    @app.cli.command("add-example-user")
    def add_example_user():
        ExampleUserTable.add(
            username="admin",
            password="password",
            disabled=False,
        )

    @app.cli.command("update-example-user")
    def update_example_user():
        ExampleUserTable.update(
            user_id=1,
            username="admin-updated",
            private_key="private_key",
            disabled=False,
        )

    @app.cli.command("delete-example-user")
    def delete_example_user():
        ExampleUserTable.delete(
            user_id=1,
        )
"""

    # Format to: None
    collections_context_processors_py = """\
def collection(app: object):
    @app.context_processor
    def example__utility_processor():
        \"""
        Usage:
        {{ format_price(100.33) }} -> $100.33
        \"""

        def example__format_price(amount, currency='$'):
            return '{1}{0:.2f}'.format(amount, currency)

        return dict(format_price=example__format_price)
"""

    # Format to: None
    collections_error_handlers_py = """\
def collection(app: object):
    from flask import render_template

    @app.errorhandler(400)
    def error_400(error):
        return render_template(
            "errors/400.html",
        ), 400

    @app.errorhandler(401)
    def error_401(error):
        return render_template(
            "errors/401.html",
        ), 401

    @app.errorhandler(403)
    def error_403(error):
        return render_template(
            "errors/403.html",
        ), 403

    @app.errorhandler(404)
    def error_404(error):
        return render_template(
            "errors/404.html",
        ), 404

    @app.errorhandler(405)
    def error_405(error):
        return render_template(
            "errors/405.html",
        ), 405

    @app.errorhandler(500)
    def error_500(error):
        return render_template(
            "errors/500.html",
        ), 500
"""

    # Format to: None
    collections_filters_py = """\
def collection(app: object):
    @app.template_filter('example__num_to_month')
    def example__num_to_month(num: str) -> str:
        \"""
        Usage:
        {{ 1 | example__num_to_month }} -> January
        \"""
        if isinstance(num, int):
            num = str(num)

        months = {
            "1": "January",
            "2": "February",
            "3": "March",
            "4": "April",
            "5": "May",
            "6": "June",
            "7": "July",
            "8": "August",
            "9": "September",
            "10": "October",
            "11": "November",
            "12": "December",
        }

        if num in months:
            return months[num]
        return "Month not found"
"""

    # Format to: None
    collections_routes_py = """\
def collection(app: object):
    from flask import render_template

    @app.route("/global")
    def index():
        return render_template(
            "index.html",
        )
"""

    # Format to: name
    static_main_js = """\
console.log('This log is from the file global/static/js/main.js')
"""

    # Format to: None
    templates_extends_main_html = """\
<!doctype html>

<html lang="en">
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>Flask-BigApp</title>
    <link rel="stylesheet" href="{{ url_for('static', filename='css/water.css') }}">
    <script src="{{ url_for('static', filename='js/main.js') }}"></script>
</head>

<body>
{% include "includes/header.html" %}
{% block content %}{% endblock %}
{% include "includes/footer.html" %}
</body>
</html>
"""

    # Format to: None
    templates_includes_footer_html = """\
<hr/>
<p>This is the footer, this is located here: <code>global/templates/includes/footer.html</code></p>
<p>It is being imported in the <code>global/templates/extends/main.html</code> template.</p>
"""

    # Format to: None
    templates_includes_header_html = """\
<div style="display: flex; flex-direction: row; align-items: center; gap: 2rem; margin-bottom: 2rem;">
    <img style="border-radius: 50%"
         src="{{ url_for('static', filename='img/Flask-BigApp-Logo.svg') }}" alt="logo">
    <div>
        <h1 style="margin: 0;">Flask-BigApp</h1>
        <p>This is the header, this is located here: <code>global/templates/includes/header.html</code></p>
        <p>It is being imported in the <code>global/templates/extends/main.html</code> template.</p>
    </div>
</div>
"""

    # Format to: None
    templates_index_html = """\
{% extends 'extends/main.html' %}


{% block content %}

    <p>This is the template file located in <code>global/templates/index.html</code></p>

{% endblock %}
"""

    templates_errors_400_html = """\
<!doctype html>
<html lang="en">

<head>
    <title>400 Bad Request</title>
</head>

<body>
<p>It's not us, it's you.</p>
</body>
</html>
"""

    templates_errors_401_html = """\
<!doctype html>
<html lang="en">

<head>
    <title>401 Unauthorized</title>
</head>

<body>
<p>You lack valid authentication credentials for the requested resource</p>
</body>
</html>
"""

    templates_errors_403_html = """\
<!doctype html>
<html lang="en">

<head>
    <title>403 Forbidden</title>
</head>

<body>
<p>Access forbidden!</p>
</body>
</html>
"""

    templates_errors_404_html = """\
<!doctype html>
<html lang="en">

<head>
    <title>404 Page Not Found</title>
</head>

<body>
<p>No route associated with the URL</p>
</body>
</html>
"""

    templates_errors_405_html = """\
<!doctype html>
<html lang="en">

<head>
    <title>405 Method Not Allowed</title>
</head>

<body>
<p>Should of GET when you POST, or POST when you GET</p>
</body>
</html>
"""

    templates_errors_500_html = """\
<!doctype html>
<html lang="en">

<head>
    <title>500 Server Error!</title>
</head>

<body>
<p>There has been a server error!</p>
</body>
</html>
"""



class BlueprintFileLib:
    # Format to: NONE
    init_py = """\
from flask_bigapp import Blueprint

bp = Blueprint(__name__)

bp.import_resources("routes")


@bp.before_app_request
def before_app_request():
    bp.init_session()
"""

    # Format to: name, url_prefix
    config_toml = """\
enabled = "yes"

[settings]
url_prefix = "/{url_prefix}"
#subdomain = ""
#url_defaults = {{}}
static_folder = "static"
template_folder = "templates"
static_url_path = "/static/{name}"
#root_path = ""
#cli_group = ""

[session]
#{name}_session = "yes"
"""

    # Format to: NONE
    routes_index_py = """\
from flask import render_template

from .. import bp


@bp.route("/", methods=["GET"])
def index():
    return render_template(bp.tmpl("index.html"))
"""

    # Format to: name, location
    templates_index_html = """\
{{% extends '{name}/extends/main.html' %}}

{{% block content %}}
    <div style="display: flex; flex-direction: row; align-items: center; gap: 2rem; margin-bottom: 2rem;">
        <img style="border-radius: 50%"
             src="{{{{ url_for('{name}.static', filename='Flask-BigApp-Logo.svg') }}}}" alt="logo">
        <div>
            <h1 style="margin: 0;">{name} blueprint</h1>
            <cite>
                <h3>Here's your new blueprint.</h3>
                <p>Remember to double-check the config.toml file.
                <p/>
            </cite>
        </div>
    </div>
{{% endblock %}}
"""

    # Format to: name
    templates_extends_main_html = """\
<!doctype html>

<html lang="en">
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>Blueprint: {name}</title>
    <link rel="stylesheet" href="{{{{ url_for('{name}.static', filename='water.css') }}}}">
</head>

<body>
{{% block content %}}

{{% endblock %}}
</body>

</html>
"""


class InitAppBlueprintFileLib:
    # Format to: name, url_prefix
    config_toml = """\
enabled = "yes"

[settings]
url_prefix = "/{url_prefix}"
#subdomain = ""
#url_defaults = {{}}
#static_folder = "static"
template_folder = "templates"
#static_url_path = "/static/{name}"

[session]
#{name}_session = "yes"
"""

    templates_index_html = """\
{{% extends 'extends/main.html' %}}

{{% block content %}}
    <div style="margin-bottom: 2rem;">
        <cite>
            This is the index route of the included example blueprint.<br/>
            This template page is located in <code>blueprints/www/templates/www/index.html</code><br/>
            it extends from <code>global/templates/extends/main.html</code><br/>
            with it's route defined in <code>blueprints/www/routes/index.py</code><br/><br/>
            It's being auto imported by <code>bp.import_resources("routes")</code> <br/>
            in the <code>blueprints/www/__init__.py</code> file.
        </cite>
    </div>
{{% endblock %}}
"""



from ._global import GlobalFileLib
from .app import AppFileLib
from .blueprint import BlueprintFileLib, InitAppBlueprintFileLib

from .flask_bigapp_logo_svg import flask_bigapp_logo_svg
from .water_css import water_css

__all__ = [
    "GlobalFileLib",
    "AppFileLib",
    "BlueprintFileLib",
    "InitAppBlueprintFileLib",
    "flask_bigapp_logo_svg",
    "water_css",
]



water_css = """\
/**
 * Forced dark theme version
 */

:root {
    --background-body: #202b38;
    --background: #161f27;
    --background-alt: #1a242f;
    --selection: #1c76c5;
    --text-main: #dbdbdb;
    --text-bright: #fff;
    --text-muted: #a9b1ba;
    --links: #41adff;
    --focus: #0096bfab;
    --border: #526980;
    --code: #ffbe85;
    --animation-duration: 0.1s;
    --button-base: #0c151c;
    --button-hover: #040a0f;
    --scrollbar-thumb: var(--button-hover);
    --scrollbar-thumb-hover: rgb(0, 0, 0);
    --form-placeholder: #a9a9a9;
    --form-text: #fff;
    --variable: #d941e2;
    --highlight: #efdb43;
    --select-arrow: url("data:image/svg+xml;charset=utf-8,%3C?xml \
    version='1.0' encoding='utf-8'?%3E %3Csvg version='1.1'\
     xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3\
     .org/1999/xlink' height='62.5' width='116.9' fill='%23efef\
     ef'%3E %3Cpath d='M115.3,1.6 C113.7,0 111.1,0 109.5,1.6 \
     L58.5,52.7 L7.4,1.6 C5.8,0 3.2,0 1.6,1.6 C0,3.2 0,5.8 1.6,\
     7.4 L55.5,61.3 C56.3,62.1 57.3,62.5 58.4,62.5 C59.4,62.5\
      60.5,62.1 61.3,61.3 L115.2,7.4 C116.9,5.8 116.9,3.2 115.\
      3,1.6Z'/%3E %3C/svg%3E");
}

html {
    scrollbar-color: #040a0f #202b38;
    scrollbar-color: var(--scrollbar-thumb) var(--background-body);
    scrollbar-width: thin;
}

body {
    font-family: system-ui, -apple-system, BlinkMacSystemFont, 'Segoe UI', \
    'Roboto', 'Oxygen', 'Ubuntu', 'Cantarell', 'Fira Sans', 'Droid Sans', \
    'Helvetica Neue', 'Segoe UI Emoji', 'Apple Color Emoji', \
    'Noto Color Emoji', sans-serif;
    line-height: 1.4;
    max-width: 800px;
    margin: 20px auto;
    padding: 0 10px;
    word-wrap: break-word;
    color: #dbdbdb;
    color: var(--text-main);
    background: #202b38;
    background: var(--background-body);
    text-rendering: optimizeLegibility;
}

button {
    transition: background-color 0.1s linear,
    border-color 0.1s linear,
    color 0.1s linear,
    box-shadow 0.1s linear,
    transform 0.1s ease;
    transition: background-color var(--animation-duration) linear,
    border-color var(--animation-duration) linear,
    color var(--animation-duration) linear,
    box-shadow var(--animation-duration) linear,
    transform var(--animation-duration) ease;
}

input {
    transition: background-color 0.1s linear,
    border-color 0.1s linear,
    color 0.1s linear,
    box-shadow 0.1s linear,
    transform 0.1s ease;
    transition: background-color var(--animation-duration) linear,
    border-color var(--animation-duration) linear,
    color var(--animation-duration) linear,
    box-shadow var(--animation-duration) linear,
    transform var(--animation-duration) ease;
}

textarea {
    transition: background-color 0.1s linear,
    border-color 0.1s linear,
    color 0.1s linear,
    box-shadow 0.1s linear,
    transform 0.1s ease;
    transition: background-color var(--animation-duration) linear,
    border-color var(--animation-duration) linear,
    color var(--animation-duration) linear,
    box-shadow var(--animation-duration) linear,
    transform var(--animation-duration) ease;
}

h1 {
    font-size: 2.2em;
    margin-top: 0;
}

h1,
h2,
h3,
h4,
h5,
h6 {
    margin-bottom: 12px;
    margin-top: 24px;
}

h1 {
    color: #fff;
    color: var(--text-bright);
}

h2 {
    color: #fff;
    color: var(--text-bright);
}

h3 {
    color: #fff;
    color: var(--text-bright);
}

h4 {
    color: #fff;
    color: var(--text-bright);
}

h5 {
    color: #fff;
    color: var(--text-bright);
}

h6 {
    color: #fff;
    color: var(--text-bright);
}

strong {
    color: #fff;
    color: var(--text-bright);
}

h1,
h2,
h3,
h4,
h5,
h6,
b,
strong,
th {
    font-weight: 600;
}

q::before {
    content: none;
}

q::after {
    content: none;
}

blockquote {
    border-left: 4px solid #0096bfab;
    border-left: 4px solid var(--focus);
    margin: 1.5em 0;
    padding: 0.5em 1em;
    font-style: italic;
}

q {
    border-left: 4px solid #0096bfab;
    border-left: 4px solid var(--focus);
    margin: 1.5em 0;
    padding: 0.5em 1em;
    font-style: italic;
}

blockquote > footer {
    font-style: normal;
    border: 0;
}

blockquote cite {
    font-style: normal;
}

address {
    font-style: normal;
}

a[href^='mailto\\:']::before {
    content: '📧 ';
}

a[href^='tel\\:']::before {
    content: '📞 ';
}

a[href^='sms\\:']::before {
    content: '💬 ';
}

mark {
    background-color: #efdb43;
    background-color: var(--highlight);
    border-radius: 2px;
    padding: 0 2px 0 2px;
    color: #000;
}

a > code,
a > strong {
    color: inherit;
}

button,
select,
input[type='submit'],
input[type='reset'],
input[type='button'],
input[type='checkbox'],
input[type='range'],
input[type='radio'] {
    cursor: pointer;
}

input,
select {
    display: block;
}

[type='checkbox'],
[type='radio'] {
    display: initial;
}

input {
    color: #fff;
    color: var(--form-text);
    background-color: #161f27;
    background-color: var(--background);
    font-family: inherit;
    font-size: inherit;
    margin-right: 6px;
    margin-bottom: 6px;
    padding: 10px;
    border: none;
    border-radius: 6px;
    outline: none;
}

button {
    color: #fff;
    color: var(--form-text);
    background-color: #161f27;
    background-color: var(--background);
    font-family: inherit;
    font-size: inherit;
    margin-right: 6px;
    margin-bottom: 6px;
    padding: 10px;
    border: none;
    border-radius: 6px;
    outline: none;
}

textarea {
    color: #fff;
    color: var(--form-text);
    background-color: #161f27;
    background-color: var(--background);
    font-family: inherit;
    font-size: inherit;
    margin-right: 6px;
    margin-bottom: 6px;
    padding: 10px;
    border: none;
    border-radius: 6px;
    outline: none;
}

select {
    color: #fff;
    color: var(--form-text);
    background-color: #161f27;
    background-color: var(--background);
    font-family: inherit;
    font-size: inherit;
    margin-right: 6px;
    margin-bottom: 6px;
    padding: 10px;
    border: none;
    border-radius: 6px;
    outline: none;
}

button {
    background-color: #0c151c;
    background-color: var(--button-base);
    padding-right: 30px;
    padding-left: 30px;
}

input[type='submit'] {
    background-color: #0c151c;
    background-color: var(--button-base);
    padding-right: 30px;
    padding-left: 30px;
}

input[type='reset'] {
    background-color: #0c151c;
    background-color: var(--button-base);
    padding-right: 30px;
    padding-left: 30px;
}

input[type='button'] {
    background-color: #0c151c;
    background-color: var(--button-base);
    padding-right: 30px;
    padding-left: 30px;
}

button:hover {
    background: #040a0f;
    background: var(--button-hover);
}

input[type='submit']:hover {
    background: #040a0f;
    background: var(--button-hover);
}

input[type='reset']:hover {
    background: #040a0f;
    background: var(--button-hover);
}

input[type='button']:hover {
    background: #040a0f;
    background: var(--button-hover);
}

input[type='color'] {
    min-height: 2rem;
    padding: 8px;
    cursor: pointer;
}

input[type='checkbox'],
input[type='radio'] {
    height: 1em;
    width: 1em;
}

input[type='radio'] {
    border-radius: 100%;
}

input {
    vertical-align: top;
}

label {
    vertical-align: middle;
    margin-bottom: 4px;
    display: inline-block;
}

input:not([type='checkbox']):not([type='radio']),
input[type='range'],
select,
button,
textarea {
    -webkit-appearance: none;
}

textarea {
    display: block;
    margin-right: 0;
    box-sizing: border-box;
    resize: vertical;
}

textarea:not([cols]) {
    width: 100%;
}

textarea:not([rows]) {
    min-height: 40px;
    height: 140px;
}

select {
    background: #161f27 url("data:image/svg+xml;charset=utf-8,%\
    3C?xml version='1.0' encoding='utf-8'?%3E %3Csvg version=\
    '1.1' xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w\
    3.org/1999/xlink' height='62.5' width='116.9' fill='%23efef\
    ef'%3E %3Cpath d='M115.3,1.6 C113.7,0 111.1,0 109.5,1.6 L5\
    8.5,52.7 L7.4,1.6 C5.8,0 3.2,0 1.6,1.6 C0,3.2 0,5.8 1.6,7.\
    4 L55.5,61.3 C56.3,62.1 57.3,62.5 58.4,62.5 C59.4,62.5 60\
    .5,62.1 61.3,61.3 L115.2,7.4 C116.9,5.8 116.9,3.2 115.3,\
    1.6Z'/%3E %3C/svg%3E") calc(100% - 12px) 50% / 12px no-repeat;
    background: var(--background) var(--select-arrow) calc(100% - \
    12px) 50% / 12px no-repeat;
    padding-right: 35px;
}

select::-ms-expand {
    display: none;
}

select[multiple] {
    padding-right: 10px;
    background-image: none;
    overflow-y: auto;
}

input:focus {
    box-shadow: 0 0 0 2px #0096bfab;
    box-shadow: 0 0 0 2px var(--focus);
}

select:focus {
    box-shadow: 0 0 0 2px #0096bfab;
    box-shadow: 0 0 0 2px var(--focus);
}

button:focus {
    box-shadow: 0 0 0 2px #0096bfab;
    box-shadow: 0 0 0 2px var(--focus);
}

textarea:focus {
    box-shadow: 0 0 0 2px #0096bfab;
    box-shadow: 0 0 0 2px var(--focus);
}

input[type='checkbox']:active,
input[type='radio']:active,
input[type='submit']:active,
input[type='reset']:active,
input[type='button']:active,
input[type='range']:active,
button:active {
    transform: translateY(2px);
}

input:disabled,
select:disabled,
button:disabled,
textarea:disabled {
    cursor: not-allowed;
    opacity: 0.5;
}

::-moz-placeholder {
    color: #a9a9a9;
    color: var(--form-placeholder);
}

:-ms-input-placeholder {
    color: #a9a9a9;
    color: var(--form-placeholder);
}

::-ms-input-placeholder {
    color: #a9a9a9;
    color: var(--form-placeholder);
}

::placeholder {
    color: #a9a9a9;
    color: var(--form-placeholder);
}

fieldset {
    border: 1px #0096bfab solid;
    border: 1px var(--focus) solid;
    border-radius: 6px;
    margin: 0;
    margin-bottom: 12px;
    padding: 10px;
}

legend {
    font-size: 0.9em;
    font-weight: 600;
}

input[type='range'] {
    margin: 10px 0;
    padding: 10px 0;
    background: transparent;
}

input[type='range']:focus {
    outline: none;
}

input[type='range']::-webkit-slider-runnable-track {
    width: 100%;
    height: 9.5px;
    -webkit-transition: 0.2s;
    transition: 0.2s;
    background: #161f27;
    background: var(--background);
    border-radius: 3px;
}

input[type='range']::-webkit-slider-thumb {
    box-shadow: 0 1px 1px #000, 0 0 1px #0d0d0d;
    height: 20px;
    width: 20px;
    border-radius: 50%;
    background: #526980;
    background: var(--border);
    -webkit-appearance: none;
    margin-top: -7px;
}

input[type='range']:focus::-webkit-slider-runnable-track {
    background: #161f27;
    background: var(--background);
}

input[type='range']::-moz-range-track {
    width: 100%;
    height: 9.5px;
    -moz-transition: 0.2s;
    transition: 0.2s;
    background: #161f27;
    background: var(--background);
    border-radius: 3px;
}

input[type='range']::-moz-range-thumb {
    box-shadow: 1px 1px 1px #000, 0 0 1px #0d0d0d;
    height: 20px;
    width: 20px;
    border-radius: 50%;
    background: #526980;
    background: var(--border);
}

input[type='range']::-ms-track {
    width: 100%;
    height: 9.5px;
    background: transparent;
    border-color: transparent;
    border-width: 16px 0;
    color: transparent;
}

input[type='range']::-ms-fill-lower {
    background: #161f27;
    background: var(--background);
    border: 0.2px solid #010101;
    border-radius: 3px;
    box-shadow: 1px 1px 1px #000, 0 0 1px #0d0d0d;
}

input[type='range']::-ms-fill-upper {
    background: #161f27;
    background: var(--background);
    border: 0.2px solid #010101;
    border-radius: 3px;
    box-shadow: 1px 1px 1px #000, 0 0 1px #0d0d0d;
}

input[type='range']::-ms-thumb {
    box-shadow: 1px 1px 1px #000, 0 0 1px #0d0d0d;
    border: 1px solid #000;
    height: 20px;
    width: 20px;
    border-radius: 50%;
    background: #526980;
    background: var(--border);
}

input[type='range']:focus::-ms-fill-lower {
    background: #161f27;
    background: var(--background);
}

input[type='range']:focus::-ms-fill-upper {
    background: #161f27;
    background: var(--background);
}

a {
    text-decoration: none;
    color: #41adff;
    color: var(--links);
}

a:hover {
    text-decoration: underline;
}

code {
    background: #161f27;
    background: var(--background);
    color: #ffbe85;
    color: var(--code);
    padding: 2.5px 5px;
    border-radius: 6px;
    font-size: 1em;
}

samp {
    background: #161f27;
    background: var(--background);
    color: #ffbe85;
    color: var(--code);
    padding: 2.5px 5px;
    border-radius: 6px;
    font-size: 1em;
}

time {
    background: #161f27;
    background: var(--background);
    color: #ffbe85;
    color: var(--code);
    padding: 2.5px 5px;
    border-radius: 6px;
    font-size: 1em;
}

pre > code {
    padding: 10px;
    display: block;
    overflow-x: auto;
}

var {
    color: #d941e2;
    color: var(--variable);
    font-style: normal;
    font-family: monospace;
}

kbd {
    background: #161f27;
    background: var(--background);
    border: 1px solid #526980;
    border: 1px solid var(--border);
    border-radius: 2px;
    color: #dbdbdb;
    color: var(--text-main);
    padding: 2px 4px 2px 4px;
}

img,
video {
    max-width: 100%;
    height: auto;
}

hr {
    border: none;
    border-top: 1px solid #526980;
    border-top: 1px solid var(--border);
}

table {
    border-collapse: collapse;
    margin-bottom: 10px;
    width: 100%;
    table-layout: fixed;
}

table caption {
    text-align: left;
}

td,
th {
    padding: 6px;
    text-align: left;
    vertical-align: top;
    word-wrap: break-word;
}

thead {
    border-bottom: 1px solid #526980;
    border-bottom: 1px solid var(--border);
}

tfoot {
    border-top: 1px solid #526980;
    border-top: 1px solid var(--border);
}

tbody tr:nth-child(even) {
    background-color: #161f27;
    background-color: var(--background);
}

tbody tr:nth-child(even) button {
    background-color: #1a242f;
    background-color: var(--background-alt);
}

tbody tr:nth-child(even) button:hover {
    background-color: #202b38;
    background-color: var(--background-body);
}

::-webkit-scrollbar {
    height: 10px;
    width: 10px;
}

::-webkit-scrollbar-track {
    background: #161f27;
    background: var(--background);
    border-radius: 6px;
}

::-webkit-scrollbar-thumb {
    background: #040a0f;
    background: var(--scrollbar-thumb);
    border-radius: 6px;
}

::-webkit-scrollbar-thumb:hover {
    background: rgb(0, 0, 0);
    background: var(--scrollbar-thumb-hover);
}

::-moz-selection {
    background-color: #1c76c5;
    background-color: var(--selection);
    color: #fff;
    color: var(--text-bright);
}

::selection {
    background-color: #1c76c5;
    background-color: var(--selection);
    color: #fff;
    color: var(--text-bright);
}

details {
    display: flex;
    flex-direction: column;
    align-items: flex-start;
    background-color: #1a242f;
    background-color: var(--background-alt);
    padding: 10px 10px 0;
    margin: 1em 0;
    border-radius: 6px;
    overflow: hidden;
}

details[open] {
    padding: 10px;
}

details > :last-child {
    margin-bottom: 0;
}

details[open] summary {
    margin-bottom: 10px;
}

summary {
    display: list-item;
    background-color: #161f27;
    background-color: var(--background);
    padding: 10px;
    margin: -10px -10px 0;
    cursor: pointer;
    outline: none;
}

summary:hover,
summary:focus {
    text-decoration: underline;
}

details > :not(summary) {
    margin-top: 0;
}

summary::-webkit-details-marker {
    color: #dbdbdb;
    color: var(--text-main);
}

dialog {
    background-color: #1a242f;
    background-color: var(--background-alt);
    color: #dbdbdb;
    color: var(--text-main);
    border: none;
    border-radius: 6px;
    border-color: #526980;
    border-color: var(--border);
    padding: 10px 30px;
}

dialog > header:first-child {
    background-color: #161f27;
    background-color: var(--background);
    border-radius: 6px 6px 0 0;
    margin: -10px -30px 10px;
    padding: 10px;
    text-align: center;
}

dialog::-webkit-backdrop {
    background: #0000009c;
    -webkit-backdrop-filter: blur(4px);
    backdrop-filter: blur(4px);
}

dialog::backdrop {
    background: #0000009c;
    -webkit-backdrop-filter: blur(4px);
    backdrop-filter: blur(4px);
}

footer {
    border-top: 1px solid #526980;
    border-top: 1px solid var(--border);
    padding-top: 10px;
    color: #a9b1ba;
    color: var(--text-muted);
}

body > footer {
    margin-top: 40px;
}

@media print {
    body,
    pre,
    code,
    summary,
    details,
    button,
    input,
    textarea {
        background-color: #fff;
    }

    button,
    input,
    textarea {
        border: 1px solid #000;
    }

    body,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    pre,
    code,
    button,
    input,
    textarea,
    footer,
    summary,
    strong {
        color: #000;
    }

    summary::marker {
        color: #000;
    }

    summary::-webkit-details-marker {
        color: #000;
    }

    tbody tr:nth-child(even) {
        background-color: #f2f2f2;
    }

    a {
        color: #00f;
        text-decoration: underline;
    }
}
"""



import functools
import logging
import os
import re
import sys
import typing as t
from pathlib import Path


class Sprinkles:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


def deprecated(message: str):
    def func_wrapper(func):
        @functools.wraps(func)
        def proc_function(*args, **kwargs):
            logging.critical(f"{Sprinkles.FAIL}Function deprecated: {message}{Sprinkles.END}")
            return func(*args, **kwargs)

        return proc_function

    return func_wrapper


def if_env_replace(
        env_value: t.Optional[t.Any],
        ignore_missing_env_variables: bool = False
) -> t.Any:
    """
    Looks for the replacement pattern to swap out values in the config from_file with environment variables.
    """
    pattern = re.compile(r'<(.*?)>')

    if isinstance(env_value, str):
        if re.match(pattern, env_value):
            env_var = re.findall(pattern, env_value)[0]
            if ignore_missing_env_variables:
                return os.environ.get(env_var)
            return os.environ.get(env_var, f"{env_value} not found in environment variables")
    return env_value


def process_dict(
        this_dict: t.Optional[dict],
        key_case_switch: str = "upper",
        ignore_missing_env_variables: bool = False,
        crawl: bool = False
) -> dict:
    if this_dict is None:
        return {}

    return_dict = {}
    for key, value in this_dict.items():
        if key_case_switch == "ignore":
            cs_key = key
        else:
            cs_key = key.upper() if key_case_switch == "upper" else key.lower()

        if crawl:
            if isinstance(value, dict):
                return_dict[cs_key] = process_dict(
                    value,
                    key_case_switch,
                    ignore_missing_env_variables,
                    crawl
                )
                continue

        return_dict[cs_key] = if_env_replace(value)

    return return_dict


def cast_to_import_str(app_name: str, folder_path: Path) -> str:
    folder_parts = folder_path.parts
    parts = folder_parts[folder_parts.index(app_name):]
    if sys.version_info.major == 3:
        if sys.version_info.minor < 9:
            return ".".join(parts).replace('.py', '')
        return ".".join(parts).removesuffix('.py')
    raise NotImplementedError("Python version not supported")


def snake(value: str) -> str:
    """
    Switches name of the class CamelCase to snake_case
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', value)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def class_field(class_: str, field: str) -> str:
    """
    Switches name of the class CamelCase to snake_case and tacks on the field name

    Used for SQLAlchemy foreign key assignments

    INFO ::: This function may not produce the correct information if you are using __tablename__ in your class
    """
    return f"{snake(class_)}.{field}"


def cast_to_bool(value: t.Union[str, bool, None]) -> bool:
    if value is None:
        return False
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        true_str = ("true", "yes", "y", "1")
        false_str = ("false", "no", "n", "0")

        if value.lower() in true_str:
            return True
        elif value.lower() in false_str:
            return False
        else:
            raise TypeError(f"Cannot cast {value} to bool")
    else:
        raise TypeError(f"Cannot cast {value} to bool")



import logging
from importlib import import_module
from importlib.util import find_spec
from inspect import getmembers
from pathlib import Path
from typing import Protocol

from flask import Blueprint
from flask import session

from .helpers import init_bp_config
from .utilities import cast_to_import_str


class BigApp(Protocol):
    def import_models(
            self,
            file_or_folder: str
    ) -> None:
        ...


class BigAppBlueprint(Blueprint):
    """
    Class that extends the capabilities of the Flask Blueprint class.
    """
    enabled: bool = False
    location: Path
    bp_name: str
    package: str
    session: dict
    settings: dict

    _bigapp_instance: BigApp

    def __init__(self, dunder_name: str, config_file: str = "config.toml"):
        r"""
        :param dunder_name: Must be __name__
        :param config_file: Must be in the same directory as the blueprint, defaults to "config.toml"

        -- config.toml example:
        ::
            enabled = "yes"

            [settings]
            url_prefix = ""
            subdomain = ""
            url_defaults = { }
            static_folder = ""
            template_folder = ""
            static_url_path = ""
            #root_path = ""
            #cli_group = ""

            [session]
            var = ""
        """
        self.package = dunder_name
        self.app_name = dunder_name.split(".")[0]

        spec = find_spec(self.package)

        if spec is None:
            raise ImportError(f"Cannot find origin of {self.package}")

        self.location = Path(f"{spec.origin}").parent
        self.bp_name = self.location.name

        self.enabled, self.session, self.settings = init_bp_config(self.bp_name, self.location / config_file)

        if self.enabled:
            super().__init__(
                self.bp_name,
                self.package,
                **self.settings
            )
            self._bigapp_instance = self._set_bigapp_instance()

    def _set_bigapp_instance(self) -> BigApp:
        """
        Internal method.
        Finds the BigApp instance in the app module.
        """
        from flask_bigapp import BigApp

        app_module = import_module(self.app_name)
        for name, value in getmembers(app_module):
            if isinstance(value, BigApp):
                return value

        raise ImportError(f"Cannot find BigApp instance in {self.app_name}")

    def import_resources(self, folder: str = "routes") -> None:
        """
        Imports all the resources (cli, routes, filters, context_processors...) in the given folder.

        :param folder: Folder to look for resources in.
        Defaults to "routes".
        Must be relative.
        """
        if not self.enabled:
            return

        resource_path = self.location / folder
        if not resource_path.exists():
            raise NotADirectoryError(f"{resource_path} is not a directory")

        resources = resource_path.glob("*.py")
        for resource in resources:
            try:
                import_module(f"{self.package}.{folder}.{resource.stem}")
            except ImportError as e:
                raise ImportError(f"Error when importing {self.package}.{resource}: {e}")

    def import_nested_blueprints(self, folder: str) -> None:
        """
        Imports all blueprints in the given folder.
        Nests them under the current blueprint.
        url: /current_blueprint/nested_blueprint

        :param folder: Folder to look for nested blueprints in.
        Must be relative.
        """
        if not self.enabled:
            return

        folder_path = Path(self.location / folder)

        for potential_bp in folder_path.iterdir():
            self.import_nested_blueprint(potential_bp.as_posix())

    def import_nested_blueprint(self, blueprint: str) -> None:
        """
        Imports a single nested blueprint from the given path.

        :param blueprint: Name of the blueprint (Python Package) to import.
        Must be relative.
        """
        if not self.enabled:
            return

        if Path(blueprint).is_absolute():
            potential_bp = Path(blueprint)
        else:
            potential_bp = Path(self.location / blueprint)

        if potential_bp.exists() and potential_bp.is_dir():
            module = import_module(
                cast_to_import_str(self.package.split(".")[0], potential_bp)
            )
            for name, value in getmembers(module):
                if isinstance(value, Blueprint):
                    if hasattr(value, "enabled"):
                        if value.enabled:
                            self.register_blueprint(value)
                        else:
                            logging.debug(f"Blueprint {name} is disabled")

    def init_session(self) -> None:
        """
        Initialize the session variables found in the config from_file.
        Use this method in the before_request route.
        """
        if not self.enabled:
            return

        for key in self.session:
            if key not in session:
                session.update(self.session)
                break

    def import_models(
            self,
            file_or_folder: str
    ) -> None:
        """
        Imports all the models in the given file or folder.

        :param file_or_folder: The file or folder to import from.
        Must be relative.
        """
        if not self.enabled:
            return

        file_or_folder_path = Path(self.location / file_or_folder)

        self._bigapp_instance.import_models(file_or_folder_path.as_posix())

    def tmpl(self, template) -> str:
        """
        Pushes together the name of the blueprint and the template from_file to look for.
        This is a small-time saving method to allow you to only type
        bp.tmpl("index.html") when looking for template files.
        """
        return f"{self.name}/{template}"



import logging
import os
from pathlib import Path
import typing as t

from toml import load as toml_load

from .resources import Resources
from .utilities import cast_to_bool, process_dict


def build_database_uri(database_config_value: dict, app) -> t.Optional[str]:
    """
    Puts together the correct database URI depending on the type specified.

    Fails if type is not supported.
    """
    app_root = Path(app.root_path)

    db_dialect = database_config_value.get("DIALECT", "None")
    db_name = database_config_value.get('DATABASE_NAME', 'database')
    db_location = database_config_value.get("LOCATION", "db")
    db_port = str(database_config_value.get('PORT', 'None'))
    db_username = database_config_value.get('USERNAME', 'None')
    db_password = database_config_value.get('PASSWORD', 'None')

    allowed_dialects = ('postgresql', 'mysql', 'oracle', 'sqlite', 'mssql')

    if db_dialect == "None":
        raise ValueError("""\
Database dialect was not specified, must be: postgresql / mysql / oracle / sqlite / mssql
Example:

[DATABASE.MAIN]
ENABLED = true
DIALECT = "sqlite"
DATABASE_NAME = "database"
LOCATION = "db"
PORT = ""
USERNAME = "database"
PASSWORD = "password"

This will create a sqlite file called
database.db in a folder called db.""")

    if "sqlite" in db_dialect:
        if db_location is not None:

            if Path(db_location).exists():
                database = Path(Path(db_location) / f"{db_name}.db")
                return f"{db_dialect}:///{database}"

            db_location_path = Path(app_root / db_location)
            db_location_path.mkdir(parents=True, exist_ok=True)
            db_location_file_path = db_location_path / f"{db_name}.db"
            return f"{db_dialect}:///{db_location_file_path}"

        db_at_root = Path(app_root / f"{db_name}.db")
        return f"{db_dialect}:///{db_at_root}"

    for dialect in allowed_dialects:
        if dialect in db_dialect:
            return f"{db_dialect}://{db_username}:{db_password}@{db_location}:{db_port}/{db_name}"

    raise ValueError("""\
Database dialect is unknown, must be: postgresql / mysql / oracle / sqlite / mssql

Example:

[DATABASE.MAIN]
ENABLED = true
DIALECT = "sqlite"
DATABASE_NAME = "database"
LOCATION = "db"
PORT = ""
USERNAME = "database"
PASSWORD = "password"

This will create a sqlite file called
database.db in a folder called db.""")


def init_app_config(config_file_path: Path, ignore_missing_env_variables: bool, app) -> dict:
    """
    Processes the values from the configuration from_file.
    """
    if not config_file_path.exists():
        logging.critical("Config file was not found, creating default.config.toml to use")

        config_file_path.write_text(
            Resources.default_config.format(
                secret_key=os.urandom(24).hex())
        )

    config_suffix = ('.toml', '.tml')

    if config_file_path.suffix not in config_suffix:
        raise TypeError("Config from_file must be one of the following types: .toml / .tml")

    config = process_dict(toml_load(config_file_path))

    flask_config = process_dict(
        config.get("FLASK"),
        key_case_switch="upper",
        ignore_missing_env_variables=ignore_missing_env_variables
    )
    session_config = process_dict(
        config.get("SESSION"),
        key_case_switch="ignore",
        ignore_missing_env_variables=ignore_missing_env_variables
    )
    sqlalchemy_config = process_dict(
        config.get("SQLALCHEMY"),
        key_case_switch="upper",
        ignore_missing_env_variables=ignore_missing_env_variables,
    )
    database_config = process_dict(
        config.get("DATABASE"),
        key_case_switch="upper",
        ignore_missing_env_variables=ignore_missing_env_variables,
        crawl=True
    )

    if flask_config is not None and isinstance(flask_config, dict):
        for flask_config_key, flask_config_value in flask_config.items():
            app.config.update({flask_config_key: flask_config_value})

    if sqlalchemy_config is not None and isinstance(sqlalchemy_config, dict):
        for sqlalchemy_config_key, sqlalchemy_config_value in sqlalchemy_config.items():
            app.config.update({sqlalchemy_config_key: sqlalchemy_config_value})

    if database_config is not None and isinstance(database_config, dict):
        app.config['SQLALCHEMY_BINDS'] = dict()
        for database_config_key, database_config_values in database_config.items():
            if database_config_values.get("ENABLED", False):
                database_uri = build_database_uri(database_config_values, app)
                if database_uri:
                    if database_config_key == "MAIN":
                        app.config['SQLALCHEMY_DATABASE_URI'] = database_uri
                        continue

                    app.config['SQLALCHEMY_BINDS'].update({
                        str(database_config_key).lower(): database_uri
                    })

    return {
        "FLASK": {**flask_config, **sqlalchemy_config},
        "SESSION": session_config,
        "DATABASE": database_config
    }


def init_bp_config(blueprint_name: str, config_file_path: Path) -> tuple:
    """
    Attempts to load the and process the configuration from_file.
    """

    if not config_file_path.exists():
        raise FileNotFoundError(f"{blueprint_name} Blueprint config from_file {config_file_path.name} was not found")

    config_suffix = ('.toml', '.tml')

    if config_file_path.suffix not in config_suffix:
        raise TypeError("Config from_file must be one of the following types: .toml / .tml")

    config = process_dict(toml_load(config_file_path), key_case_switch="lower")

    enabled = cast_to_bool(config.get('enabled', False))

    if not enabled:
        return enabled, {}, {}

    session = config.get('session', {})
    settings = config.get('settings', {})

    kwargs = dict()

    valid_settings = (
        'url_prefix', 'subdomain', 'url_defaults', 'static_folder', 'template_folder', 'static_url_path', 'root_path'
    )

    for setting in valid_settings:
        if setting == 'url_prefix':
            kwargs.update(
                {'url_prefix': settings.get('url_prefix') if settings.get('url_prefix') != "" else f"/{blueprint_name}"}
            )
            continue
        if setting in settings:
            if settings.get(setting, False):
                kwargs.update({setting: settings.get(setting)})

    return enabled, session, settings



from .auth import Auth as Auth
from .auth import PasswordGeneration as PasswordGeneration
from .bigapp import BigApp as BigApp
from .blueprint import BigAppBlueprint as Blueprint

__all__ = ["Auth", "PasswordGeneration", "BigApp", "Blueprint"]



from setuptools import setup, find_packages

readme = """
![](https://raw.githubusercontent.com/CheeseCake87/Flask-BigApp/master/_assets/Flask-BigApp-v2-Small.png)

# Flask-BigApp

![Tests](https://github.com/CheeseCake87/Flask-BigApp/actions/workflows/tests.yml/badge.svg)

## What is Flask-BigApp?

Flask-BigApp's main purpose is to help simplify the importing of blueprints, resources and models.
It has a few extra features built in to help with securing pages and password authentication.

## Getting Started

### Setup.

Create a new project folder and navigate to it.

```text
# Linux
cd /path/to/project-folder

# Windows
cd C:\path\\to\project-folder
```

### Create a virtual environment and activate it.

**Linux / MacOS**

```bash
python3 -m venv venv
```

```bash
source venv/bin/activate
```

**Windows**

```bash
python -m venv venv
```

```bash
.\\venv\Scripts\\activate
```

### Install Flask-BigApp

```bash
pip install flask-bigapp
```

### Create a new project.

```bash
flask-bigapp init
```
"""

setup(
    name='Flask-BigApp',
    version='2.3.4',
    url='https://github.com/CheeseCake87/Flask-BigApp',
    license='GNU Lesser General Public License v2.1',
    author='David Carmichael',
    author_email='carmichaelits@gmail.com',
    description='A Flask auto importer that allows your Flask apps to grow big.',
    long_description=f'{readme}',
    long_description_content_type='text/markdown',
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment', 'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent', 'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Natural Language :: English',
    ],
    install_requires=[
        'Flask',
        'Flask-SQLAlchemy',
        'toml',
    ],
    zip_safe=False,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'flask-bigapp = flask_bigapp_cli:cli',
        ]
    }
)



def test_general_setup(client):
    response = client.get('/tests/')
    assert response.status_code == 200


def test_global_route_in_folder(client):
    response = client.get('/')
    assert response.status_code == 200


def test_global_root_file_current_app(client):
    response = client.get('/current-app-import')
    assert response.status_code == 200


def test_global_root_file_factory(client):
    response = client.get('/collection-factory-import')
    assert response.status_code == 200


def test_regular_blueprint(client):
    response = client.get('/regular-blueprint')
    assert response.status_code == 200


def test_nested_blueprint(client):
    response = client.get('/tests/nested_test/')
    assert response.status_code == 200


def test_group_nested_blueprint_one(client):
    response = client.get('/tests/nested_test_one/')
    assert response.status_code == 200


def test_group_nested_blueprint_two(client):
    response = client.get('/tests/nested_test_two/')
    assert response.status_code == 200


def test_error_404(client):
    response = client.get('/tests/error-page-404')
    assert b"No route associated with the URL" in response.data


def test_error_500(client):
    response = client.get('/tests/error-page-500')
    assert b"There has been a server error!" in response.data


def test_context_processors(client):
    response = client.get('/tests/context-processors')
    assert b"$100.00" in response.data


def test_static(client):
    response = client.get('/tests/static')
    assert b"/img/Flask-BigApp-Logo.svg" in response.data


def test_database_creation(client):
    response = client.get('/tests/database-creation')
    assert b"Database created." in response.data


def test_database_population(client):
    response = client.get('/tests/database-population')
    assert b"David created" in response.data


def test_security_login_bool(client):
    client.get('/tests/login/bool')
    response = client.get('/tests/must-be-logged-in/bool')
    assert b"Security" in response.data


def test_security_login_int(client):
    client.get('/tests/login/int')
    response = client.get('/tests/must-be-logged-in/int')
    assert b"Security" in response.data


def test_security_login_str(client):
    client.get('/tests/login/str')
    response = client.get('/tests/must-be-logged-in/str')
    assert b"Security" in response.data


def test_security_login_fail(client):
    client.get('/tests/logout')
    response = client.get('/tests/must-be-logged-in/bool', follow_redirects=True)
    assert b"Login failed" in response.data


def test_permission_list(client):
    client.get('/tests/set-permission/list')
    response = client.get('/tests/must-have-permissions/std')
    assert b"Security" in response.data


def test_permission_value(client):
    client.get('/tests/set-permission/value')
    response = client.get('/tests/must-have-permissions/std')
    assert b"Security" in response.data


def test_permission_fail(client):
    response = client.get('/tests/must-have-permissions/adv', follow_redirects=True)
    assert b"Permission failed" in response.data



from pathlib import Path

import click

from .helpers import Sprinkles as Sp
from .init_new_app_blueprint import init_new_app_blueprint as _init_new_app_blueprint

from .filelib._global import GlobalFileLib
from .filelib.app import AppFileLib
from .filelib.flask_bigapp_logo_svg import flask_bigapp_logo_svg
from .filelib.water_css import water_css


def init_new_app(name):
    cwd = Path.cwd()

    app_folder = cwd / name

    if app_folder.exists():
        click.echo(f"{Sp.FAIL}{name} folder already exists!{Sp.END}")
        click.confirm("Are you sure you want to continue?", abort=True)

    gc_folder = app_folder / "global"
    models_folder = app_folder / "models"
    blueprints_folder = app_folder / "blueprints"
    extensions_folder = app_folder / "extensions"

    app_folders = (
        app_folder,
        gc_folder,
        models_folder,
        blueprints_folder,
        extensions_folder
    )

    # Create app folders
    for folder in app_folders:
        if not folder.exists():
            folder.mkdir(parents=True)

    global_tlf = (
        "cli",
        "routes",
        "templates",
        "static",
        "context_processors",
        "error_handlers",
        "filters",
    )

    global_static_folders = (
        "css",
        "js",
        "img",
    )

    global_templates_folders = (
        "errors",
        "extends",
        "includes",
    )

    global_tlf_lookup = dict()

    # Prepare global folder structure
    for folder in global_tlf:
        global_tlf_lookup[folder] = gc_folder / folder
        this_folder = gc_folder / folder
        if not this_folder.exists():
            click.echo(f"{Sp.OKGREEN}Global collections folder: {this_folder.name}, created{Sp.END}")
            this_folder.mkdir(parents=True)

    # Prepare global static folders
    for folder in global_static_folders:
        this_folder = gc_folder / "static" / folder
        if not this_folder.exists():
            click.echo(
                f"{Sp.OKGREEN}Global collections static folder: {this_folder.name}, created{Sp.END}")
            this_folder.mkdir(parents=True)

    # Prepare global templates folders
    for folder in global_templates_folders:
        this_folder = gc_folder / "templates" / folder
        if not this_folder.exists():
            click.echo(
                f"{Sp.OKGREEN}Global collections templates folder: {this_folder.name}, created{Sp.END}")
            this_folder.mkdir(parents=True)

    app_files_lu = {
        "default.config.toml": (
            app_folder / "default.config.toml",
            AppFileLib.default_config_toml,
            {}
        ),
        "__init__.py": (
            app_folder / "__init__.py",
            AppFileLib.init_py,
            {"app_name": name}
        ),
    }

    # Model files lookup
    model_files_lu = {
        "__init__.py": (
            models_folder / "__init__.py",
            AppFileLib.models_init_py,
            {"app_name": name}
        ),
        "example__table.py": (
            models_folder / "example__table.py",
            AppFileLib.models_example_table_py,
            {}
        )
    }

    # Extensions files lookup
    extensions_files_lu = {
        "__init__.py": (
            extensions_folder / "__init__.py",
            AppFileLib.extensions_init_py,
            {}
        ),
    }

    global_file_lu = {
        "cli.py": (
            global_tlf_lookup["cli"] / "cli.py",
            GlobalFileLib.collections_cli_py,
            {"app_name": name}
        ),
        "context_processors.py": (
            global_tlf_lookup["context_processors"] / "context_processors.py",
            GlobalFileLib.collections_context_processors_py,
            {}
        ),
        "error_handlers.py": (
            global_tlf_lookup["error_handlers"] / "error_handlers.py",
            GlobalFileLib.collections_error_handlers_py,
            {}
        ),
        "filters.py": (
            global_tlf_lookup["filters"] / "filters.py",
            GlobalFileLib.collections_filters_py,
            {}
        ),
        "routes.py": (
            global_tlf_lookup["routes"] / "routes.py",
            GlobalFileLib.collections_routes_py,
            {}
        ),
        "water.css": (
            global_tlf_lookup["static"] / "css" / "water.css",
            water_css,
            {}
        ),
        "main.js": (
            global_tlf_lookup["static"] / "js" / "main.js",
            GlobalFileLib.static_main_js,
            {}
        ),
        "Flask-BigApp-Logo.svg": (
            global_tlf_lookup["static"] / "img" / "Flask-BigApp-Logo.svg",
            flask_bigapp_logo_svg,
            {}
        ),
        "index.html": (
            global_tlf_lookup["templates"] / "index.html",
            GlobalFileLib.templates_index_html,
            {}
        ),
        "400.html": (
            global_tlf_lookup["templates"] / "errors" / "400.html",
            GlobalFileLib.templates_errors_400_html,
            {}
        ),
        "401.html": (
            global_tlf_lookup["templates"] / "errors" / "401.html",
            GlobalFileLib.templates_errors_401_html,
            {}
        ),
        "403.html": (
            global_tlf_lookup["templates"] / "errors" / "403.html",
            GlobalFileLib.templates_errors_403_html,
            {}
        ),
        "404.html": (
            global_tlf_lookup["templates"] / "errors" / "404.html",
            GlobalFileLib.templates_errors_404_html,
            {}
        ),
        "405.html": (
            global_tlf_lookup["templates"] / "errors" / "405.html",
            GlobalFileLib.templates_errors_405_html,
            {}
        ),
        "500.html": (
            global_tlf_lookup["templates"] / "errors" / "500.html",
            GlobalFileLib.templates_errors_500_html,
            {}
        ),
        "main.html": (
            global_tlf_lookup["templates"] / "extends" / "main.html",
            GlobalFileLib.templates_extends_main_html,
            {}
        ),
        "footer.html": (
            global_tlf_lookup["templates"] / "includes" / "footer.html",
            GlobalFileLib.templates_includes_footer_html,
            {}
        ),
        "header.html": (
            global_tlf_lookup["templates"] / "includes" / "header.html",
            GlobalFileLib.templates_includes_header_html,
            {}
        ),
    }

    # Create app files
    for file_name, (file_path, file_content, formatter) in app_files_lu.items():
        if not file_path.exists():
            if formatter:
                file_path.write_text(file_content.format(**formatter), encoding="utf-8")
            else:
                file_path.write_text(file_content, encoding="utf-8")
            click.echo(f"{Sp.OKGREEN}App file: {file_name}, created{Sp.END}")
        else:
            click.echo(f"{Sp.WARNING}App file already exists: {file_path}, skipping{Sp.END}")

    # Create model files
    for file_name, (file_path, file_content, formatter) in model_files_lu.items():
        if not file_path.exists():
            if formatter:
                file_path.write_text(file_content.format(**formatter), encoding="utf-8")
            else:
                file_path.write_text(file_content, encoding="utf-8")
            click.echo(f"{Sp.OKGREEN}Model file: {file_name}, created{Sp.END}")
        else:
            click.echo(f"{Sp.WARNING}Model file already exists: {file_name}, skipping{Sp.END}")

    # Create extensions files
    for file_name, (file_path, file_content, formatter) in extensions_files_lu.items():
        if not file_path.exists():
            if formatter:
                file_path.write_text(file_content.format(**formatter), encoding="utf-8")
            else:
                file_path.write_text(file_content, encoding="utf-8")
            click.echo(f"{Sp.OKGREEN}Extensions file: {file_name}, created{Sp.END}")
        else:
            click.echo(f"{Sp.WARNING}Extensions file already exists: {file_path}, skipping{Sp.END}")

    # Prepare global folder files
    for file_name, (file_path, file_content, formatter) in global_file_lu.items():
        if not file_path.exists():
            if formatter:
                file_path.write_text(file_content.format(**formatter), encoding="utf-8")
            else:
                file_path.write_text(file_content, encoding="utf-8")
            click.echo(f"{Sp.OKGREEN}Global collections file: {file_name}, created{Sp.END}")
        else:
            click.echo(
                f"{Sp.WARNING}Global collections file already exists: {file_path}, skipping{Sp.END}")

    _init_new_app_blueprint(f"{name}/blueprints", "www")

    click.echo(" ")
    click.echo(f"{Sp.OKBLUE}==================={Sp.END}")
    click.echo(f"{Sp.OKBLUE}Flask app deployed!{Sp.END}")
    click.echo(f"{Sp.OKBLUE}==================={Sp.END}")
    click.echo(" ")
    click.echo(f"{Sp.OKGREEN}'/' route is set by the blueprint named www{Sp.END}")
    click.echo(f"{Sp.OKGREEN}found in the blueprints folder. It is encouraged{Sp.END}")
    click.echo(f"{Sp.OKGREEN}to use blueprints to set all app routes.{Sp.END}")
    click.echo(" ")
    click.echo(f"{Sp.OKGREEN}All app (non-blueprint) resources can be found{Sp.END}")
    click.echo(f"{Sp.OKGREEN}in the global folder. Have a look through this{Sp.END}")
    click.echo(f"{Sp.OKGREEN}folder to find out more.{Sp.END}")
    click.echo(" ")
    if name == 'app':
        click.echo(f"{Sp.OKBLUE}Your app has the default name of 'app'{Sp.END}")
        click.echo(f"{Sp.OKBLUE}Flask will automatically look for this!{Sp.END}")
        click.echo(f"{Sp.OKBLUE}Run: flask run --debug{Sp.END}")
    else:
        click.echo(f"{Sp.OKBLUE}Your app has the name of '{name}'{Sp.END}")
        click.echo(f"{Sp.OKBLUE}Run: flask --app {name} run --debug{Sp.END}")
